%% Exercise 11, Tasks 11.6 - 11.12 (explicit MPC)
clear
close all

%% Task 11.6
% Here we formulate an MPC problem for the system in Exercise 7.

% MPC data
A = [0.7115 -0.4345; 0.4345 0.8853];
B = [0.2173; 0.0573];
R = 1;
Q = 10*eye(2);
[~, P] = dlqr(A, B, Q, R);
umin = -5;
umax = 5;
xmin = [-2.8; 0];
xmax = [10; 10];
N = 5; % prediction horizon

% Prediction model
model = LTISystem('A', A, 'B', B); % x(k+1) = A*x(k) + B*u(k)
model.u.min = umin;
model.u.max = umax;
model.x.min = xmin;
model.x.max = xmax;
model.u.penalty = QuadFunction(R); % quadratic penalty u'*R*u
model.x.penalty = QuadFunction(Q); % quadratic penalty x'*Q*x
% the terminal penalty is optional and has to be enabled by the user
model.x.with('terminalPenalty');
model.x.terminalPenalty = QuadFunction(P);

% Construct the explicit solution
empc = EMPCController(model, N);

%% Task 11.7
% Evaluate the explicit representation of the MPC feedback law for one
% particular initial condition
x0 = [0; 6];
% Obtain the optimal control action
uopt = empc.evaluate(x0)
% Obtain the full open-loop optimal control sequence
Uopt = empc.optimizer.feval(x0, 'primal')

%% Task 11.8
% Plot the critical regions, the feasible set and the piecewise affine
% feedback law

% Plot the critical regions
figure; empc.partition.plot(); title('Critical regions');
% Plot the feasible set
figure; empc.partition.Domain.plot(); title('Feasible set');
% Plot the feedback law
figure; 
empc.feedback.fplot(); 
xlabel('x1'); ylabel('x2'); zlabel('u0opt');
title('Feedback law');

%% Task 11.9
% Closed-loop simulation with the explicit MPC controller

loop = ClosedLoop(empc, model);
% Simulate the closed loop for 20 steps
x0 = [0; 6]; Nsim = 20;
data = loop.simulate(x0, Nsim)
figure
subplot(2, 1, 1); plot(0:Nsim, data.X); xlabel('t'); ylabel('x1, x2');
subplot(2, 1, 2); stairs(0:Nsim-1, data.U); xlabel('t'); ylabel('u');

%% Calculate a Lyapunov function to certify closed-loop stability 
lyap = loop.toSystem.lyapunov('pwq'); % PWQ = piecewise quadratic
% Successful construction means the closed-loop system is stable for all
% feasible initial conditions

%% Task 11.10
% Closed-loop simulation with a different simulation model
An = [0.6404 -0.4345; 0.4345 0.8853];
simmodel = LTISystem('A', An, 'B', B);
loop2 = ClosedLoop(empc, simmodel);
data2 = loop2.simulate(x0, Nsim)
figure
subplot(2, 1, 1); plot(0:Nsim, data2.X); xlabel('t'); ylabel('x1, x2');
subplot(2, 1, 2); stairs(0:Nsim-1, data.U); xlabel('t'); ylabel('u');

%% Task 11.11
% Tracking of non-zero references

% State reference to be tracked
xref = [0; 2];
% Enable tracking of non-zero references
model.x.with('reference');
model.x.reference = xref;
% Create a new controller
empc = EMPCController(model, N);
% Simulate the closed loop
loop = ClosedLoop(empc, model);
data = loop.simulate(x0, Nsim)
% Notice a steady-state offset
figure
subplot(2, 1, 1); 
plot(0:Nsim, data.X); 
hold on
plot(0:Nsim, repmat(xref, 1, Nsim+1), 'k--');
xlabel('t'); ylabel('x1, x2');
subplot(2, 1, 2); stairs(0:Nsim-1, data.U); xlabel('t'); ylabel('u');

%% Task 11.12
% Offset-free tracking of non-zero references

% Calculate the steady-state control input corresponding to xref
uref = B\(eye(2)-A)*xref;
% Enable tracking of non-zero references
model.u.with('reference');
model.u.reference = uref;
% Create a new controller
empc = EMPCController(model, N);
% Simulate the closed loop
loop = ClosedLoop(empc, model);
data = loop.simulate(x0, Nsim)
% Notice no steady-state offset
figure
subplot(2, 1, 1); 
plot(0:Nsim, data.X); 
hold on
plot(0:Nsim, repmat(xref, 1, Nsim+1), 'k--');
xlabel('t'); ylabel('x1, x2');
subplot(2, 1, 2); stairs(0:Nsim-1, data.U); xlabel('t'); ylabel('u');
