%% Exercise 11, Tasks 11.1 - 11.5 (parametric linear program)

%% Problem setup

% Decision variables
sdpvar x1 x2 r1 r2

% Objective function
obj = x1+6*x2;
    
% Constraints
con = [ 0 <= x1 <= 200; ...
    0 <= x2 <= r1; ...
    x1+x2 <= r2; ...
    0 <= r1 <= 500; ...
    0 <= r2 <= 700 ];

%% Task 11.1 (Numerical solution)
% Notice the minus in front of the objective function since we want to
% maximize it
solvesdp(con+[ r1==300; r2==400 ], -obj);
xopt_numerical = value([x1; x2])

%% Task 11.2 (Parametric solution)
% Here we solve the problem parametrically to obtain x1opt and x2opt as
% functions of the resources r1 and r2.

% Vector of free parameters
params = [r1; r2];
% Vector of decision variables
decs = [x1; x2];
% Let YALMIP solve the parametric linear program via MPT
ysol = solvemp(con, -obj, [], params, decs);
% Convert the YALMIP solution to the MPT format
msol = mpt_mpsol2pu(ysol);

%% Task 11.3 (Evaluation of the parametric solution)

% Vector of parameters to evaluate the solution at
r = [300; 400];
% Obtain the primal optimizer for a particular value of the parameters
xopt_parametric = msol.feval(r, 'primal')

%% Task 11.4 (Plotting of the parametric solution)

% Plot the critical regions
figure; msol.plot();
% Plot the functions x1(r1, r2) and x2(r1, r2)
figure; 
subplot(2, 1, 1); msol.fplot('primal', 'position', 1); 
xlabel('r1'); ylabel('r2'); zlabel('x1');
subplot(2, 1, 2); msol.fplot('primal', 'position', 2);
xlabel('r1'); ylabel('r2'); zlabel('x2');
% In each critical region, the optimizer is an affine function of the
% parameters, i.e., xopt = F_i*params+g_i if params \in R_i
F = {}; g = {};
for i = 1:msol.Num
    % iterate through each region and extract the primal optimizer
    F{i} = msol.Set(i).Functions('primal').F;
    g{i} = msol.Set(i).Functions('primal').g;
end

%% Task 11.5
% Decision variables
sdpvar x1 x2 r1 r2 r3
% Objective function
obj = x1+6*x2;
% Constraints
con = [ 0 <= x1 <= r3; ...
    0 <= x2 <= r1; ...
    x1+x2 <= r2; ...
    0 <= r1 <= 500; ...
    0 <= r2 <= 700; ...
    0 <= r3 <= 300 ];
% Vector of free parameters
params = [r1; r2; r3];
% Vector of decision variables
decs = [x1; x2];
% Let YALMIP solve the parametric linear program via MPT
ysol3d = solvemp(con, -obj, [], params, decs);
% Convert the YALMIP solution to the MPT format
msol3d = mpt_mpsol2pu(ysol3d);
% Plot the critical regions
figure; msol3d.plot();
% Evaluate the parametric solution for r1=200, r2=400, r3=300
xopt = msol3d.feval([200; 400; 300], 'primal')
